from abc import ABC, abstractmethod
from dataclasses import dataclass

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class IPersistenceConnector(ABC):
    @abstractmethod
    def execute(self):
        raise NotImplementedError



# coding=utf-8
# Copyright (C) 2021+ Alice, Vigo, Spain

"""Public API of Alice Petisco Framework"""
from typing import List, Type

from petisco.base.application.application import Application
from petisco.base.application.application_configurer import ApplicationConfigurer
from petisco.base.application.controller.controller import Controller
from petisco.base.application.controller.http_error import HttpError
from petisco.base.application.dependency_injection.container import Container
from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.application.dependency_injection.injector import (  # deprecated
    Injector,
)
from petisco.base.application.middleware.middleware import Middleware
from petisco.base.application.middleware.notifier_middleware import NotifierMiddleware
from petisco.base.application.middleware.print_middleware import PrintMiddleware
from petisco.base.application.notifier.not_implemented_notifier import (
    NotImplementedNotifier,
)
from petisco.base.application.notifier.notifier import Notifier
from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.base.application.notifier.notifier_message import NotifierMessage
from petisco.base.application.patterns.app_service import AppService
from petisco.base.application.patterns.repository import Repository
from petisco.base.application.use_case.use_case import UseCase
from petisco.base.application.use_case.use_case_uncontrolled_error import (
    UseCaseUncontrolledError,
)
from petisco.base.domain.errors.default_http_error_map import DEFAULT_HTTP_ERROR_MAP
from petisco.base.domain.errors.defaults.already_exists import (
    AggregateAlreadyExistError,
    AlreadyExists,
    ClientAlreadyExists,
    UserAlreadyExists,
)
from petisco.base.domain.errors.defaults.invalid_uuid import InvalidUuid
from petisco.base.domain.errors.defaults.invalid_value_object import InvalidValueObject
from petisco.base.domain.errors.defaults.not_allowed import NotAllowed
from petisco.base.domain.errors.defaults.not_found import (
    AggregateNotFoundError,
    AggregatesNotFoundError,
    ClientNotFound,
    NotFound,
    UserNotFound,
)
from petisco.base.domain.errors.domain_error import DomainError
from petisco.base.domain.errors.unknown_error import UnknownError
from petisco.base.domain.message.all_message_subscriber import AllMessageSubscriber
from petisco.base.domain.message.command import Command
from petisco.base.domain.message.command_bus import CommandBus
from petisco.base.domain.message.command_subscriber import CommandSubscriber
from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.domain_event_bus import DomainEventBus
from petisco.base.domain.message.domain_event_subscriber import DomainEventSubscriber
from petisco.base.domain.message.message import Message
from petisco.base.domain.message.message_bus import MessageBus
from petisco.base.domain.message.message_consumer import MessageConsumer
from petisco.base.domain.message.message_subscriber import MessageSubscriber
from petisco.base.domain.message.not_implemented_command_bus import (
    NotImplementedCommandBus,
)
from petisco.base.domain.message.not_implemented_domain_event_bus import (
    NotImplementedDomainEventBus,
)
from petisco.base.domain.message.not_implemented_message_bus import (
    NotImplementedMessageBus,
)
from petisco.base.domain.message.not_implemented_message_comsumer import (
    NotImplementedMessageConsumer,
)
from petisco.base.domain.message.not_implemented_message_configurer import (
    NotImplementedMessageConfigurer,
)
from petisco.base.domain.model.aggregate_root import AggregateRoot
from petisco.base.domain.model.uuid import Uuid
from petisco.base.domain.model.value_object import ValueObject
from petisco.base.domain.persistence.persistence import Persistence
from petisco.base.domain.persistence.persistence_models import PersistenceModels
from petisco.base.misc.builder import Builder
from petisco.base.misc.interface import Interface
from petisco.base.misc.result_mapper import ResultMapper
from petisco.base.testing.assert_http import assert_http

base = [
    "Controller",
    "ResultMapper",
    "Middleware",
    "NotifierMiddleware",
    "Notifier",
    "NotifierMessage",
    "NotifierExceptionMessage",
    "NotImplementedNotifier",
    "PrintMiddleware",
    "UseCase",
    "UseCaseUncontrolledError",
    "assert_http",
    "Dependency",
    "Container",
    "Message",
    "DomainEvent",
    "Command",
    "MessageBus",
    "NotImplementedMessageBus",
    "NotImplementedMessageConsumer",
    "NotImplementedMessageConfigurer",
    "DomainEventBus",
    "CommandBus",
    "NotImplementedDomainEventBus",
    "NotImplementedCommandBus",
    "MessageSubscriber",
    "MessageConsumer",
    "AggregateRoot",
    "ValueObject",
    "Uuid",
    "Interface",
    "DomainEventSubscriber",
    "CommandSubscriber",
    "AllMessageSubscriber",
    "Persistence",
    "PersistenceModels",
    "Repository",
    "AppService",
    "Application",
    "ApplicationConfigurer",
    "Builder",
]

errors = [
    "DEFAULT_HTTP_ERROR_MAP",
    "DomainError",
    "NotFound",
    "AlreadyExists",
    "AggregateNotFoundError",
    "AggregatesNotFoundError",
    "AggregateAlreadyExistError",
    "UnknownError",
    "ClientNotFound",
    "ClientAlreadyExists",
    "UserNotFound",
    "UserAlreadyExists",
    "NotAllowed",
    "InvalidUuid",
    "InvalidValueObject",
    "HttpError",
]


deprecated = ["Injector"]
modules: List[str] = []

__all__ = base + errors + deprecated + modules



from multiprocessing.pool import ThreadPool
from typing import Callable, List, Tuple


class Executable:
    def __init__(self, func: Callable, args: Tuple):
        self.func = func
        self.args = args


class PoolExecutor:
    def __init__(self, executables: List[Executable], use_multi_threading: bool = True):
        self.executables = executables
        self.use_multi_threading = use_multi_threading

    def execute(self) -> List:
        if len(self.executables) < 1:
            return []

        func_results = []
        if self.use_multi_threading:
            pool = ThreadPool(len(self.executables))
            func_results = pool.map(self._coroutine, self.executables)
            pool.close()
            pool.join()
            pool.terminate()
        else:
            for executable in self.executables:
                func_results.append(self._coroutine(executable))
        return func_results

    def _coroutine(self, executable: Executable):
        return executable.func(*executable.args)

    def clear(self):
        del self.executables
        self.executables = []



from petisco.base.application.controller.http_error import HttpError
from petisco.base.domain.errors.defaults.already_exists import (
    AggregateAlreadyExistError,
    AlreadyExists,
    ClientAlreadyExists,
    UserAlreadyExists,
)
from petisco.base.domain.errors.defaults.invalid_uuid import InvalidUuid
from petisco.base.domain.errors.defaults.invalid_value_object import InvalidValueObject
from petisco.base.domain.errors.defaults.not_found import (
    AggregateNotFoundError,
    AggregatesNotFoundError,
    ClientNotFound,
    NotFound,
    UserNotFound,
)
from petisco.base.domain.errors.domain_error import DomainError

DEFAULT_HTTP_ERROR_MAP = {
    DomainError: HttpError(status_code=500),
    NotFound: HttpError(status_code=404),
    AlreadyExists: HttpError(status_code=409),
    AggregateNotFoundError: HttpError(status_code=404),
    AggregatesNotFoundError: HttpError(status_code=404),
    AggregateAlreadyExistError: HttpError(status_code=409),
    ClientNotFound: HttpError(status_code=404),
    ClientAlreadyExists: HttpError(status_code=409),
    UserNotFound: HttpError(status_code=404),
    UserAlreadyExists: HttpError(status_code=409),
    InvalidUuid: HttpError(status_code=422),
    InvalidValueObject: HttpError(status_code=500),
}



from meiga import BoolResult, Error, Failure, isSuccess

from petisco import Uuid
from petisco.base.application.patterns.repository import Repository
from petisco.base.domain.errors.defaults.already_exists import (
    AggregateAlreadyExistError,
)
from petisco.base.domain.errors.defaults.not_found import (
    AggregateNotFoundError,
    AggregatesNotFoundError,
)


class BaseSqlRepository(Repository):
    @classmethod
    def fail_if_aggregate_already_exist(
        cls, model, aggregate_id: Uuid, result_error: Error = None
    ) -> BoolResult:
        if model:
            error = (
                AggregateAlreadyExistError(
                    cls.__name__, model.__tablename__, aggregate_id
                )
                if not result_error
                else result_error
            )
            return Failure(error)
        return isSuccess

    @classmethod
    def fail_if_aggregate_not_found(
        cls, model, aggregate_id: Uuid, result_error: Error = None
    ) -> BoolResult:
        if not model:
            error = (
                AggregateNotFoundError(cls.__name__, aggregate_id)
                if not result_error
                else result_error
            )
            return Failure(error)
        return isSuccess

    @classmethod
    def fail_if_aggregates_not_found(
        cls, model, result_error: Error = None
    ) -> BoolResult:
        if not model:
            error = (
                AggregatesNotFoundError(cls.__name__)
                if not result_error
                else result_error
            )
            return Failure(error)
        return isSuccess



from meiga import Result

from petisco.base.application.middleware.middleware import Middleware


class PrintMiddleware(Middleware):
    def before(self):
        print(
            f"{self.wrapped_class_name} -> Start | Params {dict(self.wrapped_class_input_arguments)}"
        )

    def after(self, result: Result):
        print(f"{self.wrapped_class_name} -> End | {result}")



from abc import ABC
from typing import List

from pydantic import BaseModel, PrivateAttr

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.model.uuid import Uuid


class AggregateRoot(ABC, BaseModel):
    aggregate_id: Uuid
    aggregate_version: int = 1
    _domain_events = PrivateAttr([])

    def record(self, domain_event: DomainEvent):
        self._domain_events.append(domain_event)

    def clear_domain_events(self):
        self._domain_events = []

    def pull_domain_events(self) -> List[DomainEvent]:
        return self._domain_events

    def pull_first_domain_event(self) -> DomainEvent:
        return self._domain_events[0] if len(self._domain_events) > 0 else None

    def pull_last_domain_event(self) -> DomainEvent:
        return self._domain_events[-1] if len(self._domain_events) > 0 else None



from petisco.base.domain.message.message import Message
from petisco.base.domain.message.message_bus import MessageBus


class NotImplementedMessageBus(MessageBus):
    def publish(self, message: Message):
        self._check_is_message(message)
        meta = self.get_configured_meta()
        _ = message.update_meta(meta)

    def retry_publish_only_on_store_queue(self, message: Message):
        self._check_is_message(message)
        meta = self.get_configured_meta()
        _ = message.update_meta(meta)

    def close(self):
        pass



from petisco.extra.sqlalchemy.is_sqlalchemy_available import is_sqlalchemy_available

sqlalchemy = []
if is_sqlalchemy_available():
    from petisco.extra.sqlalchemy.sql.mysql.mysql_connection import MySqlConnection
    from petisco.extra.sqlalchemy.sql.mysql.mysql_database import MySqlDatabase
    from petisco.extra.sqlalchemy.sql.sql_executor import SqlExecutor
    from petisco.extra.sqlalchemy.sql.sqlite.sqlite_connection import SqliteConnection
    from petisco.extra.sqlalchemy.sql.sqlite.sqlite_database import SqliteDatabase

    __all__ = [
        "MySqlConnection",
        "MySqlDatabase",
        "SqliteConnection",
        "SqliteDatabase",
        "SqlExecutor",
    ]
else:
    __all__ = []



from typing import Callable

from fastapi import FastAPI

from petisco.base.application.application import Application


class FastApiApplication(Application):
    fastapi_configurer: Callable[[], FastAPI]

    def get_app(self) -> FastAPI:
        return self.fastapi_configurer()



import re
from abc import abstractmethod
from types import FunctionType
from typing import List, Type

from meiga import BoolResult

from petisco.base.domain.message.command_bus import CommandBus
from petisco.base.domain.message.domain_event_bus import DomainEventBus
from petisco.base.domain.message.message import Message
from petisco.base.domain.message.message_subscriber_info import MessageSubscriberInfo
from petisco.base.domain.message.not_implemented_message_bus import (
    NotImplementedMessageBus,
)
from petisco.base.misc.interface import Interface
from petisco.base.misc.result_mapper import ResultMapper
from petisco.base.misc.wrapper import wrapper


class MetaMessageSubscriber(type, Interface):
    domain_event_bus: DomainEventBus = NotImplementedMessageBus()
    command_bus: CommandBus = NotImplementedMessageBus()
    middlewares: List = {}

    def __new__(mcs, name, bases, namespace):
        config = namespace.get("Config")

        if "handle" not in namespace:
            raise NotImplementedError(
                "Petisco MessageSubscriber must implement an handle method"
            )

        mapper = ResultMapper()

        new_namespace = {}
        for attributeName, attribute in namespace.items():
            if isinstance(attribute, FunctionType) and attribute.__name__ == "handle":
                attribute = wrapper(attribute, name, config, mapper)
            new_namespace[attributeName] = attribute

        return super().__new__(mcs, name, bases, new_namespace)

    @abstractmethod
    def subscribed_to(self) -> List[Type[Message]]:
        raise NotImplementedError()

    @abstractmethod
    def handle(self, message: Message) -> BoolResult:
        raise NotImplementedError()


class MessageSubscriber(metaclass=MetaMessageSubscriber):
    @abstractmethod
    def subscribed_to(self) -> List[Type[Message]]:
        raise NotImplementedError()

    @abstractmethod
    def handle(self, message: Message) -> BoolResult:
        raise NotImplementedError()

    @classmethod
    def __repr__(cls):
        subscriptions = cls.subscribed_to(cls)
        if not isinstance(subscriptions, list):
            subscriptions = [subscriptions]
        return f"{cls.__name__}: subscribed_to {[class_type.__name__ for class_type in subscriptions]}"

    def set_domain_event_bus(self, domain_event_bus: DomainEventBus):
        self.domain_event_bus = domain_event_bus

    def set_command_bus(self, command_bus: CommandBus):
        self.command_bus = command_bus

    def get_subscriber_name(self) -> str:
        return re.sub(r"(?<!^)(?=[A-Z])", "_", self.__class__.__name__).lower()

    def get_message_subscribers_info(self) -> List[MessageSubscriberInfo]:
        subscribers_info = []
        for class_type in self.subscribed_to():
            subscribers_info.append(MessageSubscriberInfo.from_class_type(class_type))
        return subscribers_info



from petisco.base.domain.persistence.persistence import Persistence
from petisco.extra.elastic.elastic_database import ElasticDatabase


def elastic_is_running_locally() -> bool:
    try:
        database = ElasticDatabase.local_connection_checker()
        Persistence().add(database)
        Persistence().create()
        is_running_locally = Persistence.is_available(database.name)
    except:  # noqa: E722
        is_running_locally = False
    Persistence().remove("test", skip_if_not_exist=True)
    return is_running_locally



from abc import abstractmethod

from petisco.base.domain.message.command import Command
from petisco.base.domain.message.message_bus import MessageBus


class CommandBus(MessageBus):
    @abstractmethod
    def dispatch(self, command: Command):
        raise NotImplementedError

    def publish(self, command: Command):
        self.dispatch(command)

    def retry_publish_only_on_store_queue(self, command: Command):
        pass

    def _check_is_command(self, command: Command):
        if not command or not issubclass(command.__class__, Command):
            raise TypeError("CommandBus only publishes DomainEvent objects")

    @abstractmethod
    def close(self):
        raise NotImplementedError



import os
import time

from pika import BlockingConnection, ConnectionParameters, PlainCredentials
from pika.adapters.blocking_connection import BlockingChannel
from pika.exceptions import StreamLostError

from petisco.base.misc.singleton import Singleton


class RabbitMqConnector(metaclass=Singleton):
    def __init__(self):
        self.heartbeat = os.environ.get("RABBITMQ_HEARTBEAT", 60)
        self.user = os.environ.get("RABBITMQ_USER", "guest")
        self.password = os.environ.get("RABBITMQ_PASSWORD", "guest")
        self.host = os.environ.get("RABBITMQ_HOST", "localhost")
        self.port = os.environ.get("RABBITMQ_PORT", "5672")
        self.num_max_retries_connection = os.environ.get(
            "RABBITMQ_CONNECTION_NUM_MAX_RETRIES", 15
        )
        self.wait_seconds_retry = os.environ.get(
            "RABBITMQ_CONNECTION_WAIT_SECONDS_RETRY", 1
        )
        self.open_connections = {}

    def close(self, key_connection: str):
        if self.get_connection(key_connection).is_open:
            self.get_connection(key_connection).close()

    def get_connection(self, key_connection: str):
        connection = self.open_connections.get(key_connection)

        if not connection or not connection.is_open:
            connection = self._create_connection(key_connection)

        return connection

    def get_channel(self, key_connection: str) -> BlockingChannel:
        connection = self.get_connection(key_connection)
        try:
            channel = connection.channel()
        except StreamLostError:
            connection = self.get_connection(key_connection)
            channel = connection.channel()
        return channel

    def _create_connection(self, key_connection: str):

        connection = self._create_blocking_connection(key_connection)

        self._wait_for_open_connection(connection, key_connection)

        if not connection.is_open:
            time_elapsed = round(
                (self.wait_seconds_retry * self.num_max_retries_connection), 2
            )
            raise ConnectionError(
                f"RabbitMQConnector: Impossible to obtain a open connection with host {self.host}. "
                f"Retried {self.num_max_retries_connection} in ~{time_elapsed} seconds"
            )

        return connection

    def _create_blocking_connection(self, connection_name: str):
        try:
            connection = BlockingConnection(
                ConnectionParameters(
                    heartbeat=self.heartbeat,
                    host=self.host,
                    port=int(self.port),
                    credentials=PlainCredentials(
                        username=self.user, password=self.password
                    ),
                    client_properties={"connection_name": connection_name},
                )
            )
        except:  # noqa E722
            raise ConnectionError(
                f"RabbitMQConnector: Impossible to connect to host {self.host}. "
                f"Review the following envars: [RABBITMQ_USER, RABBITMQ_PASSWORD, RABBITMQ_HOST, RABBITMQ_PORT, RABBITMQ_CONNECTION_NUM_MAX_RETRIES, RABBITMQ_CONNECTION_WAIT_SECONDS_RETRY]"
            )
        return connection

    def _wait_for_open_connection(self, connection, key_connection):
        for i in range(self.num_max_retries_connection):
            if connection.is_open:
                self.open_connections[key_connection] = connection
                break
            time.sleep(self.wait_seconds_retry)



from petisco.base.domain.errors.domain_error import DomainError
from petisco.base.domain.model.uuid import Uuid


class NotFound(DomainError):
    ...


class ClientNotFound(DomainError):
    ...


class UserNotFound(DomainError):
    ...


class AggregateNotFoundError(DomainError):
    def __init__(self, repository_name: str = None, aggregate_id: Uuid = None):
        aggregate_id_str = f" ({aggregate_id.value})" if aggregate_id else ""
        repository_str = f" (repository: {repository_name})" if repository_name else ""
        message = f"Aggregate{aggregate_id_str} not found{repository_str}"
        uuid_value = aggregate_id.value if aggregate_id else None
        super().__init__(uuid_value=uuid_value, additional_info={"message": message})


class AggregatesNotFoundError(DomainError):
    def __init__(self, repository_name: str = None):
        repository_str = f" (repository: {repository_name})" if repository_name else ""
        message = f"Aggregates not found{repository_str}]"
        super().__init__(additional_info={"message": message})
